"""
Django user objects and user managers.
"""

from ..users.userModel import standardUser

class djangoUser(standardUser):
    """Django wrapper for Krypton User
    """
    @property
    def is_authenticated(self) -> bool:
        """Returns the user's authentication state

        Returns:
            True is authenticated, False otherwise
        """
        return self.loggedin



"""
Load up databases, and configure OSSL FIPS module.
"""
# pylint: disable=cyclic-import
# pylint: disable=invalid-name

import os
import sys
import pathlib
import ctypes
import subprocess
import importlib.metadata
from sqlalchemy import DateTime, Text, create_engine, Column, Integer,\
    LargeBinary, select, Boolean
from sqlalchemy.orm import declarative_base, Session, sessionmaker

__version__ = importlib.metadata.version("krptn")

print("Hey there! Welcome from Krptn. We are setting up some things for you. "
    "In case you run into any problems, please read our "
    "common issues guide: https://docs.krptn.dev/README-FAQ.html. "
    "It is more complete than you think!")

SITE_PACKAGE = pathlib.Path(__file__).parent.parent.as_posix()
USER_DIR = pathlib.Path.home()

OPENSSL_INSTALL_PREFIX = os.environ.get("KR_OPENSSL_INSTALL", "kr-openssl-install")
OPENSSL_INSTALL_DIR = os.path.join(SITE_PACKAGE, OPENSSL_INSTALL_PREFIX)
OPENSSL_CONFIG = pathlib.Path(SITE_PACKAGE, "kr-openssl-config").as_posix()
OPENSSL_CONFIG_FILE = pathlib.Path(OPENSSL_CONFIG, "openssl.cnf").as_posix()
OPENSSL_BIN = os.path.join(OPENSSL_INSTALL_DIR, "bin")
OPENSSL_EXE = os.path.join(OPENSSL_BIN, "openssl.exe"
                           if sys.platform == "win32" else "openssl")
OSSL_LIB = os.path.join(OPENSSL_INSTALL_DIR, "lib")
if not pathlib.Path(OSSL_LIB).exists() and sys.platform == "linux":
    OSSL_LIB = os.path.join(OPENSSL_INSTALL_DIR, "lib64")
OPENSSL_MODULES = os.path.join(OSSL_LIB, "ossl-modules")
OPENSSL_FIPS_MODULE = os.path.join(OPENSSL_MODULES, "fips.dll" 
    if sys.platform == "win32"
    else ("fips.so" if sys.platform == "linux" else "fips.dylib"))
OPENSSL_FIPS_CONF = os.path.join(OPENSSL_CONFIG, "fipsmodule.cnf")

KR_DATA = pathlib.Path(pathlib.Path.home(), ".krptn-data/")
if not KR_DATA.exists():
    os.mkdir(KR_DATA.as_posix())
try:
    os.remove(OPENSSL_CONFIG_FILE)
except FileNotFoundError:
    pass

if sys.platform == "linux":
    subprocess.call(["/sbin/ldconfig", OSSL_LIB])
subprocess.call([OPENSSL_EXE, "fipsinstall", "-out", OPENSSL_FIPS_CONF,
    "-module", OPENSSL_FIPS_MODULE])

OSSL_CONF = f"""
config_diagnostics = 1
openssl_conf = openssl_init

.include {pathlib.Path(OPENSSL_FIPS_CONF).as_posix()}

[openssl_init]
providers = provider_sect

[provider_sect]
fips = fips_sect
base = base_sect

[base_sect]
activate = 1
"""

with open(OPENSSL_CONFIG_FILE, "w") as file:
    file.write(OSSL_CONF)

# Alone, it will never find these
if sys.platform == "win32":
    os.add_dll_directory(SITE_PACKAGE)
    os.add_dll_directory(OPENSSL_BIN)
    os.add_dll_directory(OPENSSL_MODULES)
elif sys.platform == "linux":
    ctypes.CDLL(os.path.join(OSSL_LIB, "libcrypto.so.3"))
elif sys.platform == "darwin":
    ctypes.CDLL(os.path.join(OSSL_LIB, "libcrypto.dylib"))

Base = declarative_base()

class DBschemas(): # pylint: disable=too-few-public-methods
    """Database Schema"""
    class CryptoTable(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        - id: int
        - ctext: bytes
        - salt: bytes
        - cipher: str
        - saltIter: int"""
        __tablename__="crypto"
        id = Column(Integer, primary_key=True)
        ctext = Column(LargeBinary)
        salt = Column(LargeBinary)
        cipher = Column(Text)
        saltIter = Column(Integer)

    class KeysTable(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        id: int
        name: str
        key: bytes
        salt: bytes
        cipher: str
        saltIter: int
        year: int"""
        __tablename__ = "keys"
        id = Column(Integer, primary_key=True)
        name = Column(Text(450), index=True)
        key = Column(LargeBinary)
        salt = Column(LargeBinary)
        cipher = Column(Text)
        saltIter = Column(Integer)
        year = Column(Integer)

    class PubKeyTable(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        Uid: int
        key: str
        krVersion: str"""
        __tablename__ = "pubKeys"
        id = Column(Integer, primary_key=True)
        Uid = Column(Integer, index=True)
        krVersion = Column(Text, default=__version__)
        key = Column(Text)

    class UserTable(Base): # pylint: disable=too-few-public-methods
        """Database
        id: int
        name: str
        pwdAuthToken: bytes,
        salt: bytes,
        mfa: bytes,
        fidoPub: bytes,
        fidoID: bytes"""
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)
        name = Column(Text(450), index=True)
        pwdAuthToken = Column(Text)
        mfa = Column(LargeBinary, default=b"*")
        fidoPub = Column(LargeBinary, default=b"*")
        fidoID = Column(LargeBinary, default=b"*")
        fidoChallenge = Column(LargeBinary, default=b"*")

    class SessionKeys(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        id: int
        Uid: int
        key: str
        exp: DateTime
        iss: DateTime"""
        __tablename__ = "sessions"
        id = Column(Integer, primary_key=True)
        Uid = Column(Integer, index=True)
        key = Column(LargeBinary)
        exp = Column(DateTime, index=True)
        iss = Column(DateTime)

    class UserData(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        Uid: int
        name: str
        value: bytes
        shared: int, default=0"""
        __tablename__ = "userData"
        id = Column(Integer, primary_key=True)
        Uid = Column(Integer, index=True)
        name = Column(Text(450), index=True)
        value = Column(LargeBinary)
    
    class UserShareTable(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        sender: int
        name: bytes
        salt: bytes
        value: bytes
        shareUid: int"""
        __tablename__ = "userShareData"
        id = Column(Integer, primary_key=True)
        sender = Column(Integer, index=True)
        name = Column(Text(450), index=True)
        salt = Column(LargeBinary)
        value = Column(LargeBinary)
        shareUid = Column(Integer, index=True)

    class PWDReset(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        Uid: int,
        key: bytes,
        iter: int,
        salt: bytes
        """
        __tablename__ = "pwdReset"
        id = Column(Integer, primary_key=True)
        Uid = Column(Integer, index=True)
        key = Column(LargeBinary)
        iter = Column(Integer)
        salt = Column(LargeBinary)

    class Logs(Base): # pylint: disable=too-few-public-methods
        """Database Schema
        logId: int,
        time: DaeTime,
        exp: DateTime,
        success: bool
        userId: int
        """
        __tablename__ = "logs"
        logId = Column(Integer, primary_key=True)
        time = Column(DateTime)
        exp = Column(DateTime)
        success = Column(Boolean)
        userId = Column(Integer, index=True)

class ConfigTemp():
    """Configuration templates"""
    defaultAlgorithm = "AES256GCM"
    APP_NAME = "KryptonApp"
    HOST_NAME = ""
    ORIGIN = ""
    defaultIterations = 800000
    defaultPasswordResetIterations = 850000
    defaultCryptoperiod = 2
    defaultSessionPeriod = 15 # Minutes
    defaultLogRetentionPeriod = 43200 # Minutes
    _cryptoDB:sessionmaker = None
    _cryptoDbEngine = None
    _altKeyDB:sessionmaker = None
    _altKeyDbEngine = None
    _userDB:sessionmaker = None
    _userDbEngine = None
    @property
    def SQLDefaultCryptoDBpath(self) -> Session:
        """
            Connection to the default database used with Crypto Class
        """
        return self._cryptoDB
    @SQLDefaultCryptoDBpath.setter
    def SQLDefaultCryptoDBpath(self, path:str) -> None:
        """
            Connection to the default database used with Crypto Class
        """
        engine = create_engine(path, echo=False, future=True)
        c = Session(engine)
        Base.metadata.create_all(engine)
        error = False
        stmt = select(DBschemas.CryptoTable)\
            .where(DBschemas.CryptoTable.id == 1)
        test = None
        try:
            test = c.scalar(stmt)
        except:
            error = True
        if test is None or error:
            stmt = DBschemas.CryptoTable(
                id = 1,
                ctext = b"Position Reserved",
                salt = b"Position Reserved",
                cipher = "None",
                saltIter = 0
            )
            c.add(stmt)
        c.autoflush = True
        c.flush()
        c.commit()
        c.close()
        self._cryptoDbEngine = engine
        self._cryptoDB = sessionmaker(engine, autoflush=True)

    @property
    def SQLDefaultKeyDBpath(self) -> Session:
        """
            Connection to the default database used by the KMS
        """
        return self._altKeyDB
    @SQLDefaultKeyDBpath.setter
    def SQLDefaultKeyDBpath(self, path:str):
        """
            Connection to the default database used by the KMS
        """
        engine = create_engine(path, echo=False, future=True)
        c = Session(engine)
        Base.metadata.create_all(engine)
        c.autoflush = True
        c.commit()
        c.flush()
        c.close()
        self._altKeyDbEngine = engine
        self._altKeyDB = sessionmaker(engine, autoflush=True)

    @property
    def SQLDefaultUserDBpath(self) -> Session:
        """
            Connection to the default database used to store User Data.
        """
        return self._userDB
    @SQLDefaultUserDBpath.setter
    def SQLDefaultUserDBpath(self, path:str):
        """
            Connection to the default database used to store User Data.
        """
        engine = create_engine(path, echo=False, future=True)
        c = Session(engine)
        Base.metadata.create_all(engine)
        error = False

        stmt = select(DBschemas.UserTable)\
            .where(DBschemas.UserTable.id == 1) # Has to be one because of shared
        test = None
        try:
            test = c.scalar(stmt)
        except:
            error = True
        if test is None or error:
            stmt = DBschemas.UserTable(
                id = 1,
                name = "Position Reserved",
                pwdAuthToken = b"Position Reserved"
            )
            c.add(stmt)
        c.autoflush = True
        c.flush()
        c.commit()
        c.close()
        self._userDbEngine = engine
        self._userDB = sessionmaker(engine, autoflush=True)

configs = ConfigTemp()

configs.SQLDefaultCryptoDBpath = "sqlite+pysqlite:///"\
    +os.path.join(USER_DIR, ".krptn-data/crypto.db")
configs.SQLDefaultKeyDBpath = "sqlite+pysqlite:///"\
    +os.path.join(USER_DIR, ".krptn-data/altKMS.db")
configs.SQLDefaultUserDBpath = "sqlite+pysqlite:///"\
    +os.path.join(USER_DIR, ".krptn-data/users.db")

#configs.SQLDefaultUserDBpath = "mssql+pyodbc://localhost/userDB?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"


"""
Provides User Models
Note for developer's working on Krypton: this only contains user model cryptography.
"""

#pylint: disable=no-member
#pylint: disable=attribute-defined-outside-init

import datetime
import os
import pickle
from typing import ByteString
from sqlalchemy import and_, delete, select, update
from sqlalchemy.orm import scoped_session
from sqlalchemy.exc import PendingRollbackError
from .. import factors
from ... import DBschemas, configs
from ... import base
from .userModelBaseAuth import AuthUser
from .userModelMFAAuth import MFAUser
from .bases import userExistRequired, user

class standardUser(AuthUser, MFAUser, user):
    """User Model for Krypton
    Check documentation.
    """
    userName:str
    _key:bytes
    _privKey:bytes
    salt:bytes
    sessionKey:bytes
    saved:bool
    loggedin:bool
    backupKeys:list[str]
    backupAESKeys:list[bytes]

    def __init__(self, userName:str=None, userID:int=None) -> None:
        super().__init__()
        self.backupAESKeys = []
        self.backupKeys = []
        self.loggedin = False
        self.FIDORequired = False
        self.c = scoped_session(configs.SQLDefaultUserDBpath)
        self.saved = False
        if userID is None and userName is None:
            return
        if userName is not None:
            stmt = select(DBschemas.UserTable.id)\
                .where(DBschemas.UserTable.name == userName)
            self.id = self.c.scalar(stmt)
            self.userName = userName
            if self.id is None:
                return
        else: # that is userID is not None
            stmt = select(DBschemas.UserTable.name)\
                .where(DBschemas.UserTable.id == userID).limit(1)
            self.userName = self.c.scalar(stmt)
            self.id = userID
            if self.userName is None:
                return
        self.saved = True

    @userExistRequired
    def setData(self, name: str, value: any) -> None:
        """Store user data as a key-value pair

        Arguments:
            name -- key

            value -- value
        """
        assert isinstance(name, str)
        try:
            self.deleteData(name)
        except:
            pass
        entry = DBschemas.UserData(
            Uid = self.id,
            name = name,
            value = base.restEncrypt(value, self._key)
        )
        self.c.add(entry)
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def getData(self, name: str) -> ByteString:
        """Get value set by setData

        Arguments:
            name -- the key

        Raises:
            ValueError: if decryption fails, or if a value is not set

        Returns:
            The value
        """
        stmt = select(DBschemas.UserData.value).where(and_(
            DBschemas.UserData.name == name,
            DBschemas.UserData.Uid == self.id))
        result = self.c.scalar(stmt)
        if result is None:
            raise ValueError("No such data saved.")
        try:
            text = base.restDecrypt(result, self._key)
        except ValueError:
            pass
        for key in self.backupAESKeys:
            retry = False
            try:
                text = base.restDecrypt(result, key)
            except ValueError:
                retry = True
            if not retry:
                break
        try:
            return text
        except NameError as exc:
            raise ValueError("Unable to decrypt the cipertext") from exc

    @userExistRequired
    def deleteData(self, name:str) -> None:
        """Delete key-value pair set by setData

        Arguments:
            name -- The key to remove
        """
        assert isinstance(name, str)
        stmt = delete(DBschemas.UserData).where(and_(DBschemas.UserData.name == name,
            DBschemas.UserData.Uid == self.id))
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def decryptWithUserKey(self, 
                    data:ByteString, salt:bytes=None, sender=None) -> bytes:
        """Decrypt data with user's key

        Arguments:
            data -- Ciphertext

            salt -- Salt

        Keyword Arguments:
            sender -- If applicable sender's user name (default: {None})

        Raises:
            ValueError: if decryption fails

        Returns:
            Plaintext
        """
        # Will also need to check the backup keys if decryption fails
        if salt is None and sender is None:
            try:
                text = base.restDecrypt(data, self._key)
            except ValueError:
                pass
            for key in self.backupAESKeys:
                retry = False
                try:
                    text = base.restDecrypt(data, key)
                except ValueError:
                    retry = True
                if not retry:
                    break
            return text
        if not isinstance(sender, int):
            stmt = select(DBschemas.UserTable.id).where(
                DBschemas.UserTable.name == sender)
            Uid = self.c.scalar(stmt)
        else:
            Uid = sender
        keys = base.getSharedKey(self._privKey, Uid, salt)
        try:
            for key in keys:
                retry = False
                try:
                    text = base.restDecrypt(data, key)
                except ValueError:
                    retry = True
                base.zeromem(key)
                if not retry:
                    break
        except ValueError:
            pass
        for privKey in self.backupKeys:
            keys = base.getSharedKey(privKey, Uid, salt)
            retry = False
            for key in keys:
                try:
                    text = base.restDecrypt(data, key)
                except ValueError:
                    retry = True
                base.zeromem(key)
                if not retry:
                    break
            if not retry:
                break
        try:
            return text
        except NameError as exc:
            raise ValueError("Unable to decrypt the ciphertext") from exc

    @userExistRequired
    def encryptWithUserKey(self,
        data:ByteString,
        otherUsers:list[int]=None) -> list[tuple[str, bytes, bytes]]:
        """Encrypt data with user's key

        Arguments:
            data -- Plaintext

        Keyword Arguments:
            otherUsers -- List of user names who can decrypt it (default: {None})

        Returns:
            If otherUsers is None: ciphertext.

            If otherUsers is not None: list of tuples (check https://docs.krptn.dev/README-USER-AUTH.html#encryption).
        """
        # pylint: disable=expression-not-assigned
        if otherUsers is None:
            ctext =  base.restEncrypt(data, self._key)
            return ctext
        salts = [os.urandom(12) for i in otherUsers]
        AESKeys = []
        otherUsers = [
            self.c.scalar(
                select(DBschemas.UserTable.id).where(DBschemas.UserTable.name == name)
            )
            for name in otherUsers
        ]
        for i, Uid in enumerate(otherUsers):
            temp = base.getSharedKey(self._privKey, Uid, salts[i])
            [base.zeromem(x) for x in temp[1:]]
            AESKeys.append(temp[0])
        results = [base.restEncrypt(data, key) for key in AESKeys]
        [base.zeromem(x) for x in AESKeys]
        return list(zip(otherUsers, results, salts))

    @userExistRequired
    def shareSet(self, name:str, data:ByteString, otherUsers:list[str]) -> None:
        """Set data readable by others

        Arguments:
            name -- The "name" of the data

            data -- The data

            otherUsers -- List of usernames who should read it
        """
        assert isinstance(name, str)
        keys = self.encryptWithUserKey(data, otherUsers)
        ids = [self.c.scalar(select(DBschemas.UserTable.id)
            .where(DBschemas.UserTable.name == user))
            for user in otherUsers]
        for i, key in enumerate(keys):
            row = DBschemas.UserShareTable(
                sender = self.id,
                name = name,
                salt = key[2],
                value = key[1],
                shareUid = ids[i]
            )
            self.c.add(row)
            self.c.flush()
        self.c.commit()

    @userExistRequired
    def shareGet(self, name:str) -> bytes:
        """Get data set by shareSet

        Arguments:
            name -- The "name of the data"

        Raises:
            ValueError: if decryption fails or requested data does not exist

        Returns:
            Decrypted data
        """
        stmt = select(DBschemas.UserShareTable).where(and_(
            DBschemas.UserShareTable.name == name,
            DBschemas.UserShareTable.shareUid == self.id))
        row:DBschemas.UserShareTable = self.c.scalar(stmt)
        if row is None:
            raise ValueError("Such data does not exist.")
        return self.decryptWithUserKey(row.value, row.salt, row.sender)

    @userExistRequired
    def shareDelete(self, name:str) -> None:
        """shareDelete Delete data set by shareSet

        Arguments:
            name -- Name of the data
        """
        assert isinstance(name, str)
        self.c.execute(delete(DBschemas.UserShareTable)\
            .where(and_(DBschemas.UserShareTable.name == name,
            DBschemas.UserShareTable.sender == self.id)))
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def generateNewKeys(self, pwd:str):
        """Regenerate Encryption keys

        Arguments:
            pwd -- Password
        """
        # Generate new Keys and add old keys to backup
        backups = self.getData("_backupAESKeys")
        self.backupAESKeys:list[bytes] = pickle.loads(backups)
        base.zeromem(backups)
        self.backupAESKeys.append(self._key)

        tag = factors.password.getAuth(pwd)
        stmt = update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.name == self.userName).\
            values(pwdAuthToken = tag)
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()
        self._key = factors.password.auth(tag, pwd)

        backups = pickle.dumps(self.backupAESKeys)
        self.setData("_backupAESKeys", backups)
        base.zeromem(backups)

        keys = base.createECCKey()
        backups = self.getData("_backupKeys")
        self.backupKeys:list[bytes] = pickle.loads(backups)
        base.zeromem(backups)
        self.backupKeys.append(self._privKey)
        backups = pickle.dumps(self.backupKeys)
        self.setData("_backupKeys", backups)
        base.zeromem(backups)
        self._privKey = keys[0]
        self.pubKey = keys[1]
        self.setData("_userPrivateKey", self._privKey)
        self.setData("_userPublicKey", self.pubKey)
        row = DBschemas.PubKeyTable(
            Uid = self.id,
            key = self.pubKey
        )
        self.c.add(row)
        self.c.flush()
        self.c.commit()

        self.setData("_accountKeysCreation", str(datetime.datetime.now().year))
        self.reload()

    @userExistRequired
    def reload(self):
        """Reload encryption keys. Warning: previous keys are not purged!
        """
        _privKey = self.getData("_userPrivateKey")
        pubKey = self.getData("_userPublicKey")
        self._privKey = _privKey.decode()
        self.pubKey = pubKey.decode()
        base.zeromem(_privKey)
        keys = self.getData("_backupAESKeys")
        self.backupAESKeys = pickle.loads(keys)
        base.zeromem(keys)
        keys = self.getData("_backupKeys")
        self.backupKeys = pickle.loads(keys)
        base.zeromem(keys)

    def __del__(self):
        if self.loggedin:
            base.zeromem(self._key)
            base.zeromem(self._privKey)
            [base.zeromem(key) for key in self.backupAESKeys]
            [base.zeromem(key) for key in self.backupKeys]
            try:
                self.c.flush()
                self.c.commit()
            except PendingRollbackError:
                self.c.rollback()
            self.c.close()



'''
https://www.krptn.dev/

A user authentication and access management system (IAM) with [Zero Knowledge security](/news/zero-knowledge/).

How we achieve this?

- All Data is encrypted (any data can be request by the developer to be secured)
- Only the appropriate users' credentials can unlock the cryptosystem (this protects you from server-side attacks)
'''
# pylint: disable=cyclic-import

from ._load import *
from . import basic
from . import auth

__all__ = ['basic', 'auth']
ignore = ['__class__', '__delattr__', '__dict__', '__dir__', '__doc__',
    '__eq__', '__format__', '__ge__', '__getattribute__', '__gt__', '__hash__',
    '__init__', '__init_subclass__', '__le__', '__lt__', '__module__',
    '__ne__', '__new__', '__reduce__', '__reduce_ex__', '__repr__', '__setattr__',
    '__sizeof__', '__str__', '__subclasshook__', '__weakref__']
search = 5

print("Krptn has been initialized and loaded. In case you run into any problems, "
    "please read our "
    "common issues guide: https://docs.krptn.dev/README-FAQ.html.")



"""
Different Auth Factors available inside krypton.
"""
import os
import base64
from webauthn import (
    generate_registration_options,
    verify_registration_response,
    options_to_json,
    base64url_to_bytes,
    generate_authentication_options,
    verify_authentication_response,
)
from webauthn.helpers.structs import (
    PublicKeyCredentialDescriptor,
    RegistrationCredential,
    UserVerificationRequirement,
    AuthenticationCredential,
)
from .. import base
from .. import configs

KEY_LEN = 32

class AuthFailed(Exception):
    """
    Exception to be raised when an error occures in a user model.
    """
    def __init__(self, *args: object) -> None:
        self.message = args[0]
        super().__init__()
    def __str__(self) -> str:
        return self.message

class password:
    """
    *Note:* no need to create an object just call the methods directly.
    Simple password authentication.

    1.) Hash the password with PBKDF2 and random salt.

    2.) Decrypt the value in the table arg.

    3.) Verify that the decryption was successfully authenticated.

    4.) Return the encryption key.
    """
    @staticmethod
    def getAuth(pwd:str):
        """Generate authentication tag for later use

        Arguments:
            pwd -- Password

        Returns:
            Auth tag
        """
        salt = os.urandom(12)
        key = base.PBKDF2(pwd, salt, keylen=KEY_LEN)
        text = os.urandom(12)
        authTag = f"""{
                base.base64encode(base.restEncrypt(text, key))
            }${
                base.base64encode(salt)
            }${
                configs.defaultIterations
            }"""
        base.zeromem(key)
        return authTag

    @staticmethod
    def auth(authTag:str, pwd:str) -> bytes:
        """Authenticate against a tag

        Arguments:
            authTag -- Tag

            pwd -- Password

        Returns:
            Encryption key if success, False otherwise
        """
        splited = authTag.split("$")
        ctext = base.base64decode(splited[0])
        salt = base.base64decode(splited[1])
        iter = int(splited[2])
        key = base.PBKDF2(pwd, salt, iter, KEY_LEN)
        try:
            base.restDecrypt(ctext, key) # This raises an error if authentication fails.
        except ValueError:
            return None
        return key ## Success

class totp:
    """
    Simple TOTP authentication
    """
    @staticmethod
    def createTOTP(userName:str):
        """Create parameters for TOTP Generate

        Arguments:
            userName -- The username

        Returns:
            shared secret, base32 encoded shared secret, totp uri
        """
        secret = os.urandom(20)
        base32Secret = base64.b32encode(secret)
        return secret, base32Secret, base.createTOTPString(secret, userName)

    @staticmethod
    def verifyTOTP(secret:bytes, otp:str) -> bool:
        """Verify TOTP

        Arguments:
            secret -- The Shared secret

            otp -- The OTP

        Returns:
            True if success False otherwise
        """
        return base.verifyTOTP(secret, otp)

class fido:
    """
    FIDO authentication support.
    """
    @staticmethod
    def register(userID:int, userName:str):
        """Start FIDO auth registration process

        Arguments:
            userID -- User's ID

            userName -- The User's username

        Returns:
            registration options and registration challenge
        """
        simple_registration_options = generate_registration_options(
            rp_id=configs.HOST_NAME,
            rp_name=configs.APP_NAME,
            user_id=str(userID),
            user_name=userName,
        )
        options = options_to_json(simple_registration_options)
        return options, simple_registration_options.challenge

    @staticmethod
    def register_verification(credentials, challenge):
        """Complete registration

        Arguments:
            credentials -- The user's fido credentials, recieved from the browser

            challenge -- The expected challenge

        Raises:
            AuthError: registration failure

        Returns:
            credential id and credential public key
        """
        registration_creds = RegistrationCredential.parse_raw(credentials)
        registration_verification = verify_registration_response(
            credential=registration_creds,
            expected_challenge=challenge,
            expected_origin=configs.ORIGIN,
            expected_rp_id=configs.HOST_NAME,
            require_user_verification=True,
        )
        success = registration_verification.credential_id == base64url_to_bytes(
            registration_creds.id
        )
        if success:
            return (registration_verification.credential_id,
                registration_verification.credential_public_key)
        raise AuthFailed("Cannot create registration for FIDO")

    @staticmethod
    def authenticate(cred_id):
        """Begin user authentication

        Arguments:
            cred_id -- The user's credential's id

        Returns:
            verification options, expected challange
        """
        authentication_options = generate_authentication_options(
            rp_id=configs.HOST_NAME,
            timeout=60000,
            user_verification=UserVerificationRequirement.PREFERRED,
            allow_credentials=[PublicKeyCredentialDescriptor(id=cred_id)],
        )
        options = options_to_json(authentication_options)
        return options, authentication_options.challenge

    @staticmethod
    def authenticate_verify(challenge:bytes, credential_public_key, credentials):
        """Complete Authentication

        Arguments:
            challenge -- The expected challange from authenticate

            credential_public_key -- The user's public key

            credentials -- The credentials provided by the user

        Returns:
            True on success, False otherwise
        """
        authentication_verification = verify_authentication_response(
            credential=AuthenticationCredential.parse_raw(credentials),
            expected_challenge=challenge,
            expected_rp_id=configs.HOST_NAME,
            expected_origin=configs.ORIGIN,
            credential_public_key=credential_public_key,
            credential_current_sign_count=0,
        )
        success = authentication_verification.new_sign_count > 0
        return success



""" Contains Abstract Base Classes for user models.
You can check this to see the declarations for functions.
"""

from functools import wraps
from abc import ABCMeta, abstractmethod
from typing import ByteString
from datetime import datetime

from sqlalchemy.orm import Session

class UserError(Exception):
    """
    Exception to be raised when an error occurs in a user model.
    """
    def __init__(self, *args: object) -> None:
        self.message = args[0]
        super().__init__()
    def __str__(self) -> str:
        return self.message

def userExistRequired(func):
    """User has to be saved in order to run this function

    Arguments:
        func -- function

    Raises:
        UserError: If user is not saved

    Returns:
        inner1
    """
    @wraps(func)
    def inner1(self, *args, **kwargs):
        """Ensure user is saved

        Raises:
            UserError: If user is not saved

        Returns:
            N/A
        """
        if self.saved and self.loggedin:
            return func(self, *args, **kwargs)
        raise UserError("This user has not yet been saved or is logged out.")
    return inner1

class user(metaclass=ABCMeta):
    """Base Class for User Models.
    You can check this to see whether a method is implemented in user models.
    """
    c:Session
    @abstractmethod
    def delete(self):
        """Delete a user

        Returns:
            None
        """
    @abstractmethod
    def login(self, pwd:str, mfaToken:str=None, fido:str=None):
        """Log the user in

        Keyword Arguments:
            pwd -- Password (default: {None})

            otp -- One-Time Password (default: {None})

            fido -- Fido Token (default: {None})

        Raises:
            UserError: Password is not set

        Returns:
            Session Key, None if user is not saved
        """
    @abstractmethod
    def restoreSession(self, key:bytes):
        """Resume session from key

        Arguments:
            key -- Session Key
        """
    @abstractmethod
    def logout(self):
        """logout Logout the user and delete the current Session
        """
    @abstractmethod
    def enableMFA(self):
        """The method name says it all."""
    @abstractmethod
    def disableMFA(self):
        """The method name says it all."""
    @abstractmethod
    def saveNewUser(self, name:str, pwd:str):
        """Save a new user

        Arguments:
            name -- User Name

            pwd -- Password

        Keyword Arguments:
            fido -- Fido Token (default: {None})

        Raises:
            ValueError: If user is already saved
        """
    @abstractmethod
    def getData(self, name: str) -> ByteString:
        """Get value set by setData

        Arguments:
            name -- the key

        Raises:
            AttributeError: if a value is not set

        Returns:
            The value
        """
    @abstractmethod
    def setData(self, name: str, value: any) -> None:
        """Store user data as a key-value pair

        Arguments:
            name -- key

            value -- value
        """
    @abstractmethod
    def deleteData(self, name:str) -> None:
        """Delete key-value pair set by setData

        Arguments:
            name -- The key to remove
        """
    @abstractmethod
    def decryptWithUserKey(self,
                           data:ByteString, salt:bytes=None, sender=None) -> bytes:
        """Decrypt data with user's key

        Arguments:
            data -- Ciphertext

            salt -- Salt

        Keyword Arguments:
            sender -- If applicable sender's user name (default: {None})

        Returns:
            Plaintext
        """
    @abstractmethod
    def encryptWithUserKey(self, data:ByteString, otherUsers:list[str]) -> bytes:
        """Encrypt data with user's key

        Arguments:
            data -- Plaintext

        Keyword Arguments:
            otherUsers -- List of user nameswho can decrypt it  (default: {None})

        Returns:
            List of tuples of form (user name, ciphertext, salt), check: https://docs.krptn.dev/README-USER-AUTH.html#encryption.
        """
    @abstractmethod
    def generateNewKeys(self, pwd:str):
        """Regenerate Encryption keys

        Arguments:
            pwd -- Password
        """
    @abstractmethod
    def resetPWD(self, key:str, newPWD:str):
        """Reset Password

        Arguments:
            key -- Key as provided to enablePWDReset
        """
    @abstractmethod
    def reload(self):
        """Reload encryption keys. Warning: previous keys are not purged!
        """
    @abstractmethod
    def enablePWDReset(self):
        """Enable Password Reset

        Arguments:
            key -- The key needed to reset
        """
    @abstractmethod
    def revokeSessions(self):
        """Revoke all Sessions for this User

        Raises:
            UserError: If the user does not exist
        """
    @abstractmethod
    def shareGet(self, name:str) -> bytes:
        """Get data set by shareSet

        Arguments:
            name -- The "name of the data"

        Raises:
            ValueError: if decryption fails

        Returns:
            Decrypted data
        """
    @abstractmethod
    def shareSet(self, name:str, data:ByteString, otherUsers:list[str]) -> None:
        """Set data readable by others

        Arguments:
            name -- The "name" of the data

            data -- The data

            otherUsers -- List of usernames who should read it
        """
    @abstractmethod
    def shareDelete(self, name:str) -> None:
        """shareDelete Delete data set by shareSet

        Arguments:
            name -- Name of the data
        """
    @abstractmethod
    def logFailure(self):
        """logFailure Log a login failure
        """
    @abstractmethod
    def getLogs(self) -> list[list[datetime, bool]]:
        """getLogs Get the login logs for the user
        """


import argparse
import sys
from . import configs, Base

def cleanDatabase():
    Base.metadata.drop_all(configs._cryptoDbEngine)
    Base.metadata.drop_all(configs._altKeyDbEngine)
    Base.metadata.drop_all(configs._userDbEngine)
    print("Cleaning Database completed!")

parser = argparse.ArgumentParser(description='Krptn CLI')
parser.add_argument('--clean', dest='doAction', const=cleanDatabase, nargs='?',
    default=lambda: 'Not cleaning Krptn database',
    help='Clean Krptn\'s default database (.krptn-data)')
args = parser.parse_args(sys.argv[1:])
args.doAction()



"""
Basic security related classes.
"""
from .base import restEncrypt, restDecrypt, zeromem, PBKDF2
from datetime import datetime
import os
from typing import ByteString
from sqlalchemy import select, func
from sqlalchemy.orm import Session, scoped_session
from . import configs, base, DBschemas
SQLDefaultCryptoDBpath:Session = configs.SQLDefaultCryptoDBpath
SQLDefaultKeyDBpath:Session = configs.SQLDefaultKeyDBpath

class KeyManagementError(Exception):
    """Error in Key Management System

    For example, compliance issues

    Arguments:
        Exception -- Inherits base Exception class
    """
    def __init__(self, *args: object) -> None:
        self.message = args[0]
        super().__init__()
    def __str__(self) -> str:
        return self.message

class KMS():
    """
    They Key Management System
    """
    def _cipher(self, text:ByteString, pwd:ByteString, salt:bytes, iterations:int):
        """Encrypt a string

        Arguments:
            text -- Plain text

            pwd -- Password

            salt -- Salt for hashing

            iterations -- Iterations for hashing

        Returns:
            Cipher text
        """
        key = PBKDF2(pwd, salt, iterations) if iterations > 0 else pwd
        r = restEncrypt(text, key)
        zeromem(key)
        return r
    def _decipher(self, ctext:ByteString, pwd:ByteString, salt:bytes, iterations:int):
        """Decrypt a string

        Arguments:
            ctext -- Cipher text

            pwd -- Password

            salt -- Salt for Hashing

            iterations -- Iterations for hashing

        Returns:
            Plaintext
        """
        key = PBKDF2(pwd, salt, iterations) if iterations > 0 else pwd
        r = restDecrypt(ctext, key)
        zeromem(key)
        return r

    def __init__(self, keyDB:Session=scoped_session(SQLDefaultKeyDBpath))->None:
        """The title says it all"""
        self.c:Session = keyDB

    def getKey(self, name:str, pwd:ByteString=None, force:bool=False) -> bytes:
        """Get a Key

        Arguments:
            name -- Name of the key to get

        Keyword Arguments:
            pwd -- Password (default: {None})

            force -- Override Cryptoperiod Compliance errors (default: {False})

        Raises:
            ValueError: If the key does not exist

            KeyManagementError: If the key has expired - set force=True to override

            ValueError: If an unsupported cipher is used

            ValueError: Wrong passwords were provided or the key was tampered with

        Returns:
            The key as python bytes
        """
        stmt = select(DBschemas.KeysTable).where(DBschemas.KeysTable.name == name)\
            .limit(1)
        key:DBschemas.KeysTable = self.c.scalar(stmt)
        if key is None:
            raise ValueError("Such key does not exist")
        if datetime.now().year - key.year >= configs.defaultCryptoperiod and not force:
            raise KeyManagementError("This key has expired. Please add force to the argument to retrieve it anyway.")
        if key.cipher != configs.defaultAlgorithm:
            raise ValueError("Unsupported Cipher")
        r = self._decipher(key.key, pwd, key.salt, key.saltIter)
        splited = r.split(b"$")
        if splited[1] != name.encode() or splited[2] != str(key.year).encode():
            raise ValueError("Wrong passwords have been provided or the database has been tampered with.")
        result = base.base64decode(splited[0])
        zeromem(r)
        zeromem(splited[0])
        return result

    def createNewKey(self, name:str, pwd:ByteString=None) -> str:
        """Create a new key and store it

        Arguments:
            name -- Name of the Key

        Keyword Arguments:
            pwd -- Password (default: {None})

        Raises:
            KeyError: If key with same name already exists

        Returns:
            The key as python bytes
        """
        year = datetime.today().year
        stmt = select(DBschemas.KeysTable).where(DBschemas.KeysTable.name == "name")
        a = True
        try:
            self.c.scalars(stmt).one()
        except:
            a=False
        finally:
            if a:
                raise KeyError("Such a name already exists")
        k = os.urandom(32)
        s = os.urandom(12)
        rebased = base.base64encode(k)
        editedRebased = rebased+f"${name}${year}"
        ek = self._cipher(editedRebased, pwd, s, configs.defaultIterations)
        zeromem(rebased)
        zeromem(editedRebased)
        key = DBschemas.KeysTable(
            name = name,
            key = ek,
            salt = s,
            cipher = configs.defaultAlgorithm,
            saltIter = configs.defaultIterations,
            year = year
        )
        self.c.add(key)
        self.c.flush()
        self.c.commit()
        return k

    def removeKey(self, name:str, pwd:ByteString=None) -> None:
        """Delete a Key

        Arguments:
            name -- Name of the Key

        Keyword Arguments:
            pwd -- Password (default: {None})
        """
        zeromem(self.getKey(name, pwd, True))
        stmt = select(DBschemas.KeysTable).where(DBschemas.KeysTable.name == name)\
            .limit(1)
        key:DBschemas.KeysTable = self.c.scalar(stmt)
        self.c.delete(key)
        self.c.flush()
        self.c.commit()
        return
    
    def __del__(self):
        self.c.close()

class Crypto(KMS):
    '''
    Crypto Class (see Documentation)
    '''
    def __init__(self, keyDB:Session=scoped_session(SQLDefaultCryptoDBpath)):
        """The title says it all"""
        self.c:scoped_session = keyDB
        stmt = select(func.max(DBschemas.CryptoTable.id))
        self.num = self.c.scalar(stmt)
        super().__init__(self.c)

    def secureCreate(self, data:ByteString, pwd:ByteString=None, _num:int=None):
        """Store Encrypted Data

        Arguments:
            data -- Plaintext data

        Keyword Arguments:
            pwd -- Password To Decrypt (default: {None})

            _num -- Not good idea to set! Id to store in DB (default: {None})

        Returns:
            Integer to be passed to secureRead to return data
        """
        if _num is None:
            self.num+=1
            _num = self.num
        key = self.createNewKey(str(_num), pwd)
        salt = os.urandom(12)
        keyOb = DBschemas.CryptoTable(
            id = _num,
            ctext = self._cipher(data, key, salt, 0),
            salt = salt,
            cipher = configs.defaultAlgorithm,
            saltIter = configs.defaultIterations
        )
        self.c.add(keyOb)
        zeromem(key)
        self.c.flush()
        self.c.commit()
        return _num

    def secureRead(self, num:int, pwd:ByteString):
        """Read data from secureCreate

        Arguments:
            num -- Integer returned from secureCreate

            pwd -- Password set in secureCreate

        Returns:
            Plaintext data
        """
        stmt = select(DBschemas.CryptoTable).where(DBschemas.CryptoTable.id ==num)\
            .limit(1)
        ctext = self.c.scalar(stmt)
        reset = False
        try:
            key = self.getKey(str(num), pwd)
        except KeyManagementError:
            reset = True
            key = self.getKey(str(num), pwd, True)
        text = self._decipher(ctext.ctext, key, ctext.salt, 0)
        if reset:
            self.secureUpdate(num, text, pwd)
        zeromem(key)
        return text

    def secureUpdate(self, num:int, new:ByteString, pwd:ByteString):
        """Update Entry Set by secureCreate

        Arguments:
            num -- Integer id of entry

            new -- New data to set

            pwd -- Password
        """
        self.secureDelete(num, pwd)
        self.secureCreate(new, pwd, num)

    def secureDelete(self, num:int, pwd:ByteString=None) -> None:
        """Delete Data set by secureCreate

        Arguments:
            num -- Integer id of entry

        Keyword Arguments:
            pwd -- Password (default: {None})
        """
        zeromem(self.getKey(str(num), pwd, True))
        stmt = select(DBschemas.CryptoTable).where(DBschemas.CryptoTable.id == num)
        key:DBschemas.CryptoTable = self.c.scalar(stmt)
        self.c.delete(key)
        self.c.flush()
        self.removeKey(str(num), pwd)



""" This module contains auth functions for models
"""
#pylint: disable=no-member
#pylint: disable=attribute-defined-outside-init
#pylint: disable=abstract-method
import datetime
import os
import pickle
from sqlalchemy import delete, select, func, update
from .. import factors, _utils
from ... import DBschemas, configs
from ... import base
from .bases import userExistRequired, UserError, user

class AuthUser(user):
    """Auth Logic for User Models
    """
    def login(self, pwd:str, mfaToken:str="", fido:str=None):
        """Log the user in

        Keyword Arguments:
            pwd -- Password

            otp -- One-Time Password (default: {""})

            fido -- Fido Credentials (default: {None})

        Raises:
            UserError: Password is not set or wrong password is provided.

        Returns:
            Session Key
        """
        def _finishAuthError():
            """_finishAuthError Prevent timing attacks
            """
            factors.password.auth(base.base64encode(os.urandom(32)), pwd)
        if not self.saved:
            _finishAuthError()
            raise UserError("This user does not exist.")
        authTag:DBschemas.UserTable = self.c.scalar(
            select(DBschemas.UserTable).where(DBschemas.UserTable.id == self.id)
            .limit(1)
        )
        if authTag.fidoID != b"*":
            if fido is None or factors.fido.authenticate_verify(authTag.fidoChallenge,
                                                        authTag.fidoPub, fido) is False:
                self.FIDORequired = True
                self.logFailure()
                _finishAuthError()
                raise UserError("Failed to verify FIDO credentials.")
        if authTag.pwdAuthToken is None:
            self.logFailure()
            _finishAuthError()
            raise UserError("User must have a password set.")
        self._key = factors.password.auth(authTag.pwdAuthToken, pwd)
        if self._key is None:
            self.logFailure()
            raise UserError("Wrong password")
        if authTag.mfa != b"*":
            mfa = base.restDecrypt(authTag.mfa, self._key)
            if not factors.totp.verifyTOTP(mfa, mfaToken):
                base.zeromem(self._key)
                self.logFailure()
                # No _finishAuthError() is intentional
                raise UserError("Wrong MFA Token")
        restoreKey = os.urandom(32)
        self.sessionKey = base.restEncrypt(self._key, restoreKey)
        token = DBschemas.SessionKeys(
            Uid = self.id,
            key = self.sessionKey,
            iss = datetime.datetime.now(),
            exp = (datetime.datetime.now() +
                   datetime.timedelta(minutes=configs.defaultSessionPeriod))
        )
        self.c.add(token)
        self.c.flush()
        self.loggedin = True
        time = int(self.getData("_accountKeysCreation").decode())
        if (datetime.datetime.now().year - time) >= 2:
            self.generateNewKeys(pwd)
        self.reload()
        self.c.flush()
        encoded = base.base64encode(restoreKey)
        base.zeromem(restoreKey)
        row = DBschemas.Logs(time=datetime.datetime.now(),
            exp=(datetime.datetime.now() +
                 datetime.timedelta(minutes=configs.defaultLogRetentionPeriod)),
            success=True,
            userId = self.id)
        self.c.add(row)
        self.c.flush()
        self.c.commit()
        return encoded
    
    def logFailure(self):
        """logFailure Log a login failure
        """
        row = DBschemas.Logs(time=datetime.datetime.now(),
            exp=(datetime.datetime.now()
                 + datetime.timedelta(minutes=configs.defaultLogRetentionPeriod)),
            success=False,
            userId = self.id)
        self.c.add(row)
        self.c.flush()
        self.c.commit()
    
    @userExistRequired
    def getLogs(self):
        """getLogs Get the login logs for the user
        """
        logs:list[DBschemas.Logs] = self.c.scalars(select(DBschemas.Logs)\
            .where(DBschemas.Logs.userId == self.id)\
            .order_by(DBschemas.Logs.time.desc()))
        return [[log.time, log.success] for log in logs]

    @userExistRequired
    def logout(self):
        """logout Logout the user and delete the current Session
        """
        base.zeromem(self._key)
        base.zeromem(self._privKey)
        stmt = delete(DBschemas.SessionKeys)\
            .where(DBschemas.SessionKeys.key == self.sessionKey)
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()
        self.c.expunge_all()
        self.loggedin = False

    @userExistRequired
    def delete(self):
        """Delete a user
        """
        _utils.cleanUpSessions(self.c, self.id)
        self.c.execute(
            delete(DBschemas.UserTable).where(DBschemas.UserTable.id == self.id))
        self.c.execute(
            delete(DBschemas.PubKeyTable).where(DBschemas.PubKeyTable.Uid == self.id))
        self.c.execute(
            delete(DBschemas.UserData).where(DBschemas.UserData.Uid == self.id))
        self.c.execute(
            delete(DBschemas.UserShareTable).where(
                DBschemas.UserShareTable.sender == self.id))
        self.c.execute(
            delete(DBschemas.PWDReset).where(DBschemas.PWDReset.Uid == self.id))
        self.c.flush()
        self.c.commit()
        base.zeromem(self._key)
        base.zeromem(self._privKey)

    def restoreSession(self, key):
        """Resume sessoin from key

        Arguments:
            key -- Session Key
        """
        if key is None:
            raise UserError("Session does not exist or is expired.")
        _utils.cleanUpSessions(session=self.c)
        decodedKey = base.base64decode(key)
        stmt = select(DBschemas.SessionKeys).where(DBschemas.SessionKeys.Uid == self.id)
        rows:list[DBschemas.SessionKeys] = self.c.scalars(stmt)
        success = False
        for row in rows:
            if row is None:
                raise UserError("Session does not exist or is expired.")
            try:
                self._key = base.restDecrypt(row.key, decodedKey)
                self.sessionKey = row.key
            except ValueError:
                pass
            else:
                success = True
        if not success:
            raise UserError("Session does not exist or is expired.")
        self.loggedin = True
        self.reload()

    def saveNewUser(self, name:str, pwd:str) -> bytes:
        """Save a new user

        Arguments:
            name -- User Name

            pwd -- Password

        Raises:
            ValueError: If user is already saved
        """
        assert isinstance(name, str) # No need to assert pwd as it will be done in C++
        if len(name) >= 450:
            raise ValueError("User name must be less then 450 characters.")
        if self.saved:
            raise ValueError("This user is already saved.")
        s = self.c.scalar(select(DBschemas.UserTable).where(
            DBschemas.UserTable.name == name))
        if s is not None:
            raise ValueError("This user is already exists.")
        self.userName = name
        stmt = select(func.max(DBschemas.UserTable.id))
        self.id = self.c.scalar(stmt) + 1
        keys = base.createECCKey()
        self._privKey = keys[0]
        self.pubKey = keys[1]
        key = DBschemas.PubKeyTable(
            Uid = self.id,
            key = self.pubKey
        )
        self.c.add(key)
        self.c.flush()
        tag = factors.password.getAuth(pwd)
        userEntry = DBschemas.UserTable(
            id = self.id,
            name = name,
            pwdAuthToken = tag
        )
        self.c.add(userEntry)
        self.c.flush()
        self._key = factors.password.auth(tag, pwd)
        self.saved = True
        self.loggedin = True
        self.setData("_userPrivateKey", self._privKey)
        self.setData("_userPublicKey", self.pubKey)
        self.setData("_backupKeys", pickle.dumps([]))
        self.setData("_backupAESKeys", pickle.dumps([]))
        self.setData("_accountKeysCreation", str(datetime.datetime.now().year))
        self.c.flush()
        self.c.commit()
        return self.login(pwd=pwd)

    def revokeSessions(self):
        """Revoke all Sessions for this User

        Raises:
            UserError: If the user does not exist
        """
        if not self.saved:
            raise UserError("This user does not exist.")
        _utils.cleanUpSessions(self.c, self.id)

    @userExistRequired
    def changeUserName(self, newUserName:str):
        """changeUserName Change the user's username

        Arguments:
            newUserName -- The new username (string)
        """
        assert isinstance(newUserName, str)
        stmt = update(DBschemas.UserTable).where(DBschemas.UserTable.id == self.id).\
            values(name = newUserName)
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()
        self.userName = newUserName



""" Utils to help code
"""
import datetime
from sqlalchemy import delete
from sqlalchemy.orm import scoped_session
from .. import DBschemas

def cleanUpSessions(session:scoped_session, userID:int = None):
    """Cleanup Expired Session or Remove all sessions for a user

    Arguments:
        session -- The database session to use

    Keyword Arguments:
        userID -- The ID for which to delete all sessions (even if not expired) (default: {None})
    """
    #If userID is provided all sessions linked to it will be deleted (even if they are not expired).
    now = datetime.datetime.now()
    if userID is not None:
        session.execute(
            delete(
                DBschemas.SessionKeys
            ).where(
                DBschemas.SessionKeys.Uid == userID
            ))
        session.flush()
    session.execute(
        delete(DBschemas.SessionKeys).where(DBschemas.SessionKeys.exp <= now))
    session.execute(
        delete(DBschemas.Logs).where(DBschemas.Logs.exp <= now))
    session.flush()
    session.commit()



"""
Django middleware to add user object as needed to the request.
"""

from django.http import HttpRequest
from .users import djangoUser
from ..users.bases import UserError


def kryptonLoginMiddleware(get_response):
    """Django middleware to add user object as needed to the request.


    Arguments:
        get_response -- Function to get HttpResponse

    Returns:
        Response
    """
    def KrLoginMiddleWare(request:HttpRequest):
        """Middleware

        Arguments:
            request -- HttpRequest

        Returns:
            HttpResponse
        """
        try:
            user = djangoUser(userID = request.COOKIES["_KryptonUserID"])
            user.restoreSession(request.COOKIES["_KryptonSessionToken"])
        except UserError: return get_response(request)
        except KeyError: return get_response(request)
        request.user = user
        return get_response(request)
    return KrLoginMiddleWare



""" Extended auth logic
"""
#pylint: disable=no-member
#pylint: disable=attribute-defined-outside-init
#pylint: disable=abstract-method
import os
from sqlalchemy import select, delete, update

from .. import factors
from ... import DBschemas, configs, base
from .bases import userExistRequired, user, UserError

class MFAUser(user):
    """MFA for Krypton Users
    """
    @userExistRequired
    def enablePWDReset(self) -> list[str]:
        """Enable PWD Reset

        Returns:
            The recovery codes that unlock the account
        """
        # pylint: disable=invalid-name
        PKeys = [base.genOTP() for i in range(10)]
        self.c.execute(delete(DBschemas.PWDReset)\
                       .where(DBschemas.PWDReset.Uid == self.id))
        for PKey in PKeys:
            salt = os.urandom(32)
            key = base.PBKDF2(PKey, salt, configs.defaultPasswordResetIterations, 32)
            skey = base.restEncrypt(self._key, key)
            base.zeromem(key)
            row = DBschemas.PWDReset(
                Uid = self.id,
                key = skey,
                iter = configs.defaultPasswordResetIterations,
                salt = salt
            )
            self.c.add(row)
            self.c.flush()
        self.c.commit()
        return PKeys

    def resetPWD(self, key:str, newPWD:str):
        assert isinstance(key, str)
        assert isinstance(newPWD, str)
        """Reset a PWD using a recovery code

        Arguments:
            key -- The recovery code

            newPWD -- The new PWD

        Raises:
            ValueError: if the reset fails
        """
        rows = self.c.execute(select(DBschemas.PWDReset)\
                .where(DBschemas.PWDReset.Uid == self.id)).scalars().all()
        reset = False
        for row in rows:
            krKey = base.PBKDF2(key, row.salt, row.iter, 32)
            try:
                self._key = base.restDecrypt(row.key, krKey)
            except ValueError:
                continue
            base.zeromem(krKey)
            self.loggedin = True
            self.c.execute(delete(DBschemas.PWDReset)\
                           .where(DBschemas.PWDReset.id == row.id))
            self.c.flush()
            self.generateNewKeys(newPWD)
            token = self.login(newPWD)
            reset = True
            break
        if not reset:
            base.sleepOutOfGIL()
            raise ValueError("Password reset failure")
        self.c.commit()
        return token

    @userExistRequired
    def disablePWDReset(self):
        """Disbale PWD and revoke all codes
        """
        self.c.execute(delete(DBschemas.PWDReset)\
                       .where(DBschemas.PWDReset.Uid == self.id))
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def enableMFA(self):
        """Enable TOTP MFA

        Returns:
            base32 encoded shared secret, QR code string
        """
        secret, base32Secret, string = factors.totp.createTOTP(self.userName)
        stmt = update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.name == self.userName).\
            values(mfa = base.restEncrypt(secret, self._key))
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()
        base.zeromem(secret)
        return base32Secret, string

    @userExistRequired
    def disableMFA(self):
        """Disable TOTP based MFA
        """
        stmt = update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.name == self.userName).\
            values(mfa = b"*")
        self.c.execute(stmt)
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def beginFIDOSetup(self):
        """Being FIDO Registration
        """
        options, challenge = factors.fido.register(self.id, self.userName)
        self.setData("_tempFIDORegisterChallenge", challenge)
        return options

    @userExistRequired
    def completeFIDOSetup(self, response):
        """Finish FIDO Setup

        Arguments:
            repsonse -- The response from the client
        """
        challenge = self.getData("_tempFIDORegisterChallenge")
        self.deleteData("_tempFIDORegisterChallenge")
        credID, credKey = factors.fido.register_verification(response, challenge)
        self.c.execute(update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.name == self.userName).\
            values(fidoPub=credKey, fidoID=credID))
        self.c.flush()
        self.c.commit()

    @userExistRequired
    def removeFIDO(self):
        """Remove the FIDO Auth from Server
        """
        self.c.execute(update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.name == self.userName).\
            values(fidoPub=b"*", fidoID=b"*"))
        self.c.flush()
        self.c.commit()

    def getFIDOOptions(self):
        """Obtain FIDO options before Auth

        Returns:
            Fido Options as string, { "error": "No keys availble" } if FIDO is not setup
        """
        if not self.saved:
            raise UserError("This user does not exist.")
        stmt = select(DBschemas.UserTable).where(DBschemas.UserTable.id == self.id)\
            .limit(1)
        authTag:DBschemas.UserTable = self.c.scalar(stmt)
        if authTag.fidoID == b"*":
            return '{ "error": "No keys availble" }'
        options, challenge = factors.fido.authenticate(authTag.fidoID)
        self.c.execute(update(DBschemas.UserTable)\
            .where(DBschemas.UserTable.id == self.id).\
            values(fidoChallenge = challenge))
        self.c.flush()
        self.c.commit()
        return options



"""
Loads __CryptoLib and contains wrappers.
"""
#pylint: disable=import-error
# disbaled pylint because __CryptoLib is not built in CI/CD tests

import ctypes
import sys
import base64
import os
from typing import ByteString
from sqlalchemy import select
from sqlalchemy.orm import scoped_session, Session
try:
    import __CryptoLib
except ImportError as err:
    if sys.platform == "win32" and not os.path.isfile("C:/Windows/System32/MSVCP140.dll"):
        raise RuntimeError("This module requires Microsoft Visual C/C++ runtime. "
            "Please download it from https://learn.microsoft.com/en-US/cpp/windows/latest-supported-vc-redist."
            ) from err
    raise err
from . import configs, DBschemas, OPENSSL_CONFIG_FILE, OPENSSL_MODULES

Adrr = id

#: Load FIPS Validated resolver
__CryptoLib.fipsInit(OPENSSL_CONFIG_FILE, OPENSSL_MODULES)

#: Wrappers for __CryptoLib
#: Help static analyzers automatically figure out function arguments, returns, etc..
def restEncrypt(data:ByteString, key:bytes) -> bytes:
    """Encrypt Data for at rest storage

    Arguments:
        data -- Plain text

        key -- 32-bit key

    Returns:
        Cipher text
    """
    return __CryptoLib.AESEncrypt(data, key, len(data))

def restDecrypt(data:bytes, key:bytes) -> bytes:
    """Decrypt Data from restEncrypt

    Arguments:
        data -- Cipher text

        key -- 32-bit key

    Returns:
        Plain text
    """
    return __CryptoLib.AESDecrypt(data, key)

def base64encode(data:ByteString) -> str:
    """Base64 Encoding

    Arguments:
        data -- Text to encode

    Returns:
        Base64 encoded string
    """
    return __CryptoLib.base64encode(data, len(data))

def base64decode(data:ByteString) -> ByteString:
    """Decode base64

    Arguments:
        data -- Base64 encoded string

    Returns:
        Base64 decoded bytes
    """
    return __CryptoLib.base64decode(data, len(data))

def createECCKey() -> tuple[str, str]:
    """create an Eliptic Curve Key

    Encoded in P.E.M. format

    Returns:
        Returns a tuple like (privateKey:str, publicKey:str)
    """
    return __CryptoLib.createECCKey()

def ECDH(privKey:str, peerPubKey:str, salt:bytes, keylen:int=32) -> bytes:
    """Elliptic Curve Diffie-Helman

    Arguments:
        privKey -- P.E.M. Encoded private key

        peerPubKey -- P.E.M. Encoded public key

        salt -- Salt used for Key Derivation Function

    Keyword Arguments:
        keylen -- Len of the key (default: {32})

    Returns:
        Key as python bytes
    """
    return __CryptoLib.ECDH(privKey, peerPubKey, salt, keylen)

def getSharedKey(privKey:str, peerID:int, salt:bytes, keylen:int=32) -> list[bytes]:
    """Get users' shared key

    Get a shared key for two users using ECDH.

    Arguments:
        privKey -- User's private EC Key (in P.E.M. format)

        peerID -- Other User's ID

        salt -- Salt used for KDF

    Keyword Arguments:
        keylen -- Len of key to return (default: {32})

    Returns:
        List of keys as python bytes
    """
    assert isinstance(privKey, str)
    assert isinstance(peerID, int)
    # pylint: disable=no-member
    session:Session = scoped_session(configs.SQLDefaultUserDBpath)
    pubKeys:list[DBschemas.PubKeyTable] = session.scalars(
        select(DBschemas.PubKeyTable)
        # In future, krVersion can be used to detect compatability issues
        .where(DBschemas.PubKeyTable.Uid == peerID)
        .order_by(DBschemas.PubKeyTable.id.desc())).all()
    results = [__CryptoLib.ECDH(privKey, pubKey.key, salt, keylen) for pubKey in pubKeys]
    session.close()
    return results

def PBKDF2(text:ByteString, salt:ByteString, iterations:int=configs.defaultIterations, keylen:int=32) -> bytes:
    """PBKDF2 with SHA512

    Arguments:
        text -- Plain text
        salt -- Salt

    Keyword Arguments:
        iterations -- Iteration count (default: {configs.defaultIterations})

        keylen -- Len of key to return (default: {32})

    Returns:
        The key as python bytes
    """
    return __CryptoLib.PBKDF2(text, len(text), salt, iterations, len(salt), keylen)

def zeromem(obj:ByteString)->int:
    """Set the byte/string to \\x00

    WARNING! Improper use leads to severe memory corruption.
    Ensure you only use it with bytes and string objects.
    Also, on PyPy this function does nothing to avoid corruption.

    Arguments:
        obj -- Object to do this on (bytes and str are supported!)

    Returns:
        Result from memset.
    """
    assert (isinstance(obj, str) or isinstance(obj, bytes))
    if "PyPy" not in sys.version:
        return ctypes.memset(id(obj)+(sys.getsizeof(obj)-len(obj)),0,len(obj))
    return None

def verifyTOTP(secret:bytes, code:str) -> bool:
    """Verify a 6-digit TOTP

    Arguments:
        secret -- The shared secret

        code -- The code to verify

    Returns:
        True is success false otherwise
    """
    return __CryptoLib.totpVerify(secret, code)

def createTOTPString(secret:bytes, user:str) -> str:
    """Create a TOTP String that can be scanned by Auth Apps

    Arguments:
        secret -- The base32 encoded shared secret

    Returns:
        The String to be converted to QR code
    """
    assert isinstance(user, str)
    s = base64.b32encode(secret)
    secret = s.decode()
    stripped = secret.strip("=")
    string = f"otpauth://totp/{configs.APP_NAME}:{user}?secret={stripped}&issuer=KryptonAuth&algorithm=SHA1&digits=6&period=30"
    zeromem(s)
    zeromem(secret)
    zeromem(stripped)
    return string

def genOTP() -> str:
    """Generate an 12-digit OTP/PIN.

    Returns:
        The OTP/PIN as python string
    """
    return __CryptoLib.genOTP()

def sleepOutOfGIL(seconds:int=5) -> bool:
    """Sleep for seconds while releasing the GIL.

    Keyword Arguments:
        seconds -- Number of seconds to slepp for (default: {5})

    Returns:
        True
    """
    return __CryptoLib.sleepOutOfGIL(seconds)



import os
import sys
import pathlib
import warnings
from glob import glob
from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension

folder = pathlib.Path(__file__).parent.as_posix()

DEBUG = sys.argv.count("--debug") >= 1
OPENSSL_INSTALL_PREFIX = os.environ.get("KR_OPENSSL_INSTALL", "kr-openssl-install")

macros = []
link_libararies = ["crypto"]
runtime_libs = [os.path.join(folder, f"{OPENSSL_INSTALL_PREFIX}/lib")]
extra_args = []
library_dirs = [f"{OPENSSL_INSTALL_PREFIX}/lib"]

package_data = [
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libcrypto-3-x64.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libssl-3-x64.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libssl-3-arm64.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libcrypto-3-arm64.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libcrypto-3.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/libssl-3.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/ossl-modules/fips.dll",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/openssl.exe",
  f"../{OPENSSL_INSTALL_PREFIX}/bin/openssl",
  f"../{OPENSSL_INSTALL_PREFIX}/lib64/libcrypto.so.3",
  f"../{OPENSSL_INSTALL_PREFIX}/lib64/libssl.so.3",
  f"../{OPENSSL_INSTALL_PREFIX}/lib64/ossl-modules/fips.so",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/libcrypto.so.3",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/libssl.so.3",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/ossl-modules/fips.so",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/libcrypto.dylib",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/libssl.dylib",
  f"../{OPENSSL_INSTALL_PREFIX}/lib/ossl-modules/fips.dylib",
  "../kr-openssl-config/openssl.cnf",
  "../kr-openssl-config/fipsmodule.cnf"
]

if not pathlib.Path(folder, OPENSSL_INSTALL_PREFIX).exists():
  warnings.warn("We detected that you are likely building Krptn from source in an unsuitable manner. "
    "Do not attempt to build Krptn from source without reading https://docs.krptn.dev/README-BUILD.html first. "
    "Doing so is a terrible mistake and is likely to cause failures and other errors."
    "If you are not building Krptn from source or you don't get any errors, please ignore this false positive.", 
    RuntimeWarning, stacklevel=2)

if sys.platform == "linux":
  library_dirs += [f"{OPENSSL_INSTALL_PREFIX}/lib64"]
  runtime_libs += [os.path.join(folder, f"{OPENSSL_INSTALL_PREFIX}/lib64")]
elif sys.platform == "win32":
  link_libararies = ["libcrypto", "user32", "WS2_32", "GDI32", "ADVAPI32", "CRYPT32"]
  macros += [("WIN", None)]
  runtime_libs = []
elif sys.platform == "darwin":
  extra_args += ["-std=c++17", "-O0"] # Disable optimizationas as they trigger segementation faults

setup(
  package_data={"": package_data},
  include_package_data=True,
  packages=find_packages(),
  ext_modules=[Pybind11Extension('__CryptoLib',
    glob("CryptoLib/*.cpp"),
    include_dirs=[f"{OPENSSL_INSTALL_PREFIX}/include", "CryptoLib"],
    library_dirs=library_dirs,
    libraries=link_libararies,
    runtime_library_dirs=runtime_libs,
    extra_compile_args=extra_args,
    extra_link_args=extra_args,
    define_macros=macros)
  ],
)
